from pandas import DataFrame
from pandas.core.dtypes.common import pandas_dtype


def df_metadata(df: DataFrame) -> DataFrame:
    columns = DataFrame({'column': df.columns,
                            'index': False,
                            'datatype': [str(dt) for dt in df.dtypes]})
    indexes = DataFrame({'column': [f'_no_name_{i}' if name==None else name for i, name in enumerate(df.index.names)],
                            'index': True,
                            'datatype': str(df.index.dtype) if len(df.index.names)==1
                                            else [str(dt) for dt in df.index.dtypes]})  # type: ignore
    return indexes.append(columns).reset_index(drop=True)


def unname_no_names(df) -> None:
    if len(df.index.names)==1:
        if df.index.name == '_no_name':
            df.index.name = None
    else:
        df.index.names = [None if name==f'_no_name_{i}' else name for i, name in enumerate(df.index.names)]
        

def name_no_names(df) -> None:
    if len(df.index.names)==1:
        if df.index.name == None:
            df.index.name = '_no_name'
    else:
        df.index.names = [f'_no_name_{i}' if name==None else '_no_name_' for i, name in enumerate(df.index.names)]

        
def cast_type(series):
    return [pandas_dtype(n) for n in series]


def convert_meta_to_dict(meta: DataFrame) -> dict:
    return {col: typ for col, typ in zip(meta['column'], meta['datatype'])}


def df_cols_to_numpy(df: DataFrame) -> None:
    """Convert each column to numpy data type"""
    for name in df.columns:
        df[name] = df[name].to_numpy()


"""
Store Pandas DataFrames in a hdf5 file in a specified directory path.
Also saves the DataFrame metadata in the hdf5 file.

When saved table is loaded back into a DataFrame,
the indexes and datatypes are converted back from saved metadata.

Example
-------
import pandas as pd
import trashpandas as tp

df = pd.DataFrame({'name': ['Joe', 'Bob', 'John'], 'age': [23, 34, 44]})

# Create HdfStorage object with hdf5 file in current directory path.
# TrashPandas will create the hdf5 file if it doesn't exist.'
storage = tp.HdfStorage('data.h5')

# Store DataFrame in hdf5 table named 'people'
# and store metadata as hdf5 table named '_people_metadata'
storage.store(df, 'people') 
# or assign DataFrame to item
storage['people'] = df

# Retrieve DataFrame using HdfStorage load method.
df = storage.load('people')
# or use table name key
df = storage['people']

# Delete stored table using HdfStorage delete method.
storage.delete('people')
# or use del on table name key
del storage['people']

# Or use functions instead of HdfStorage class
tp.store_df_hdf5(df, 'people', 'data.h5')

df = tp.load_df_hdf5('people', 'data.h5')

tp.delete_table_hdf5('people', 'data.h5')
"""

import os

from h5py import File
from pandas import DataFrame, read_hdf

from trashpandas.interfaces import IStorage
from trashpandas.utils import cast_type, convert_meta_to_dict, df_metadata, name_no_names, unname_no_names
from trashpandas.utils import df_cols_to_numpy


class HdfStorage(IStorage):
    def __init__(self, hdf5_path: str) -> None:
        """Takes folder path to hdf5 file where DataFrames and metadata are stored."""
        self.path = hdf5_path
        # create hdf5 file if it doesn't exist
        create_hdf5_file(hdf5_path)

    def __repr__(self) -> str:
        return f"HdfStorage('{self.path}')"

    def __setitem__(self, key: str, other: DataFrame) -> None:
        """Store DataFrame and metadata in hdf5 file."""
        self.store(other, key)

    def __getitem__(self, key: str) -> DataFrame:
        """Retrieve DataFrame from hdf5 file."""
        return self.load(key)

    def __delitem__(self, key: str) -> None:
        """Delete DataFrame and metadata from hdf5 file."""
        self.delete(key)

    def store(self, df: DataFrame, table_name: str) -> None:
        """Store DataFrame and metadata in hdf5 file."""
        store_df_hdf5(df, table_name, self.path)

    def load(self, table_name: str) -> DataFrame:
        """Retrieve DataFrame from hdf5 file."""
        return load_df_hdf5(table_name, self.path)

    def delete(self, table_name: str) -> None:
        """Delete DataFrame and metadata from hdf5 file."""
        delete_table_hdf5(table_name, self.path)

    def load_metadata(self, table_name: str, schema=None) -> DataFrame:
        """Retrieve DataFrame metadata from hdf5 file."""
        return load_metadata_hdf5(table_name, self.path)

    def table_names(self) -> list[str]:
        """Get list of stored non-metadata table names."""
        return table_names_hdf5(self.path)

    def metadata_names(self) -> list[str]:
        """Get list of stored metadata table names."""
        return metadata_names_hdf5(self.path)


def create_hdf5_file(path: str) -> None:
    """Create hdf5 file if it doesn't exist"""
    if not os.path.exists(path):
        hf = File(path, 'w')
        hf.close()


def store_df_hdf5(df: DataFrame, table_name: str, path: str) -> None:
    """Store DataFrame and metadata in hdf5 file."""
    df = df.copy()
    name_no_names(df)
    metadata = df_metadata(df)
    df_cols_to_numpy(df)
    create_hdf5_file(path)
    df.to_hdf(path, key=table_name)
    metadata.to_hdf(path, key=f'_{table_name}_metadata', index=False)


def load_df_hdf5(table_name: str, path: str) -> DataFrame:
    """Retrieve DataFrame from hdf5 file."""
    meta_name = f'_{table_name}_metadata'

    if meta_name not in metadata_names_hdf5(path):
        return _first_load_df_hdf5(table_name, path)

    metadata = _read_cast_metadata_hdf5(meta_name, path)
    types = convert_meta_to_dict(metadata)
    indexes = list(metadata['column'][metadata['index']==True])
    df = read_hdf(path, key=table_name).reset_index().astype(types).set_index(indexes)
    unname_no_names(df)
    return df


def delete_table_hdf5(table_name: str, path: str) -> None:
    """Delete DataFrame and metadata from hdf5 file."""
    meta_name = f'_{table_name}_metadata'
    with File(path,  "a") as hf:
        del hf[table_name]
        del hf[meta_name]


def load_metadata_hdf5(table_name: str, path: str) -> DataFrame:
    """Retrieve DataFrame metadata from hdf5 file."""
    meta_name = f'_{table_name}_metadata'
    return _read_cast_metadata_hdf5(meta_name, path)


def table_names_hdf5(path: str) -> list[str]:
    """Get list of stored non-metadata table names."""
    table_names = _hdf5_keys(path)
    return [name for name in table_names
                if '_metadata' not in name and name[0]!='_']


def metadata_names_hdf5(path: str) -> list[str]:
    """Get list of stored metadata table names."""
    table_names = _hdf5_keys(path)
    return [name for name in table_names
                if '_metadata' in name and name[0]=='_']


def _hdf5_keys(path: str) -> list[str]:
    """Get list of all stored table names."""
    with File(path, 'r') as hf:
        names = list(hf.keys())
    return names


def _read_cast_metadata_hdf5(table_name: str, path: str) -> DataFrame:
    """Load metadata table and cast column datatypes column."""
    meta = read_hdf(path, key=table_name)
    meta['datatype'] = cast_type(meta['datatype'])
    return DataFrame(meta)


def _first_load_df_hdf5(table_name: str, path: str) -> DataFrame:
    """Load a sql table that has no metadata stored, create and store metadata"""
    # Load a table that has no metadata stored, create and store metadata
    df = DataFrame(read_hdf(path, key=table_name))
    store_df_hdf5(df, table_name, path)
    return df


"""
Convert tables between storage types. 

CsvStorage <-> SqlStorage
CsvStorage <-> HdfStorage
CsvStorage <-> PickleStorage
SqlStorage <-> HdfStorage
SqlStorage <-> PickleStorage
HdfStorage <-> PickleStorage

"""

from sqlalchemy.engine import Engine

from trashpandas.interfaces import IStorage
from trashpandas.csv import CsvStorage
from trashpandas.pickle import PickleStorage
from trashpandas.sql import SqlStorage
from trashpandas.hdf5 import HdfStorage


def convert_table_storage(table_name: str, starting_storage: IStorage, ending_storage: IStorage) -> None:
    df = starting_storage.load(table_name)
    ending_storage.store(df, table_name)


def convert_all_tables_storage(starting_storage: IStorage, ending_storage: IStorage) -> None:
    for table_name in starting_storage.table_names():
        convert_table_storage(table_name, starting_storage, ending_storage)


def csv_to_sql(table_name: str, csv_folder_path: str, engine: Engine) -> None:
    csv_storage = CsvStorage(csv_folder_path)
    sql_storage = SqlStorage(engine)
    convert_table_storage(table_name, csv_storage, sql_storage)


def csv_to_sql_all(csv_folder_path: str, engine: Engine) -> None:
    csv_storage = CsvStorage(csv_folder_path)
    sql_storage = SqlStorage(engine)
    convert_all_tables_storage(csv_storage, sql_storage)


def csv_to_hdf(table_name: str, csv_folder_path: str, hdf_file_path: str) -> None:
    csv_storage = CsvStorage(csv_folder_path)
    hdf_storage = HdfStorage(hdf_file_path)
    convert_table_storage(table_name, csv_storage, hdf_storage)


def csv_to_hdf_all(csv_folder_path: str, hdf_file_path: str) -> None:
    csv_storage = CsvStorage(csv_folder_path)
    hdf_storage = HdfStorage(hdf_file_path)
    convert_all_tables_storage(csv_storage, hdf_storage)


def csv_to_pickle(table_name: str, csv_folder_path: str, pickle_folder_path: str) -> None:
    csv_storage = CsvStorage(csv_folder_path)
    pickle_storage = PickleStorage(pickle_folder_path)
    convert_table_storage(table_name, csv_storage, pickle_storage)


def csv_to_pickle_all(csv_folder_path: str, pickle_folder_path: str) -> None:
    csv_storage = CsvStorage(csv_folder_path)
    pickle_storage = PickleStorage(pickle_folder_path)
    convert_all_tables_storage(csv_storage, pickle_storage)


def sql_to_csv(table_name: str, engine: Engine, csv_folder_path: str) -> None:
    sql_storage = SqlStorage(engine)
    csv_storage = CsvStorage(csv_folder_path)
    convert_table_storage(table_name, sql_storage, csv_storage)


def sql_to_csv_all(engine: Engine, csv_folder_path: str) -> None:
    sql_storage = SqlStorage(engine)
    csv_storage = CsvStorage(csv_folder_path)
    convert_all_tables_storage(sql_storage, csv_storage)


def sql_to_hdf(table_name: str, engine: Engine, hdf_file_path: str) -> None:
    sql_storage = SqlStorage(engine)
    hdf_storage = HdfStorage(hdf_file_path)
    convert_table_storage(table_name, sql_storage, hdf_storage)


def sql_to_hdf_all(engine: Engine, hdf_file_path: str) -> None:
    sql_storage = SqlStorage(engine)
    hdf_storage = HdfStorage(hdf_file_path)
    convert_all_tables_storage(sql_storage, hdf_storage)


def sql_to_pickle(table_name: str, engine: Engine, pickle_folder_path: str) -> None:
    sql_storage = SqlStorage(engine)
    pickle_storage = PickleStorage(pickle_folder_path)
    convert_table_storage(table_name, sql_storage, pickle_storage)


def sql_to_pickle_all(engine: Engine, pickle_folder_path: str) -> None:
    sql_storage = SqlStorage(engine)
    pickle_storage = PickleStorage(pickle_folder_path)
    convert_all_tables_storage(sql_storage, pickle_storage)


def hdf_to_csv(table_name: str, hdf_file_path: str, csv_folder_path: str) -> None:
    hdf_storage = HdfStorage(hdf_file_path)
    csv_storage = CsvStorage(csv_folder_path)
    convert_table_storage(table_name, hdf_storage, csv_storage)


def hdf_to_csv_all(hdf_file_path: str, csv_folder_path: str) -> None:
    hdf_storage = HdfStorage(hdf_file_path)
    csv_storage = CsvStorage(csv_folder_path)
    convert_all_tables_storage(hdf_storage, csv_storage)


def hdf_to_sql(table_name: str, hdf_file_path: str, engine: Engine) -> None:
    hdf_storage = HdfStorage(hdf_file_path)
    sql_storage = SqlStorage(engine)
    convert_table_storage(table_name, hdf_storage, sql_storage)


def hdf_to_sql_all(hdf_file_path: str, engine: Engine) -> None:
    hdf_storage = HdfStorage(hdf_file_path)
    sql_storage = SqlStorage(engine)
    convert_all_tables_storage(hdf_storage, sql_storage)


def hdf_to_pickle(table_name: str, hdf_file_path: str, pickle_folder_path: str) -> None:
    hdf_storage = HdfStorage(hdf_file_path)
    pickle_storage = PickleStorage(pickle_folder_path)
    convert_table_storage(table_name, hdf_storage, pickle_storage)


def hdf_to_pickle_all(hdf_file_path: str, pickle_folder_path: str) -> None:
    hdf_storage = HdfStorage(hdf_file_path)
    pickle_storage = PickleStorage(pickle_folder_path)
    convert_all_tables_storage(hdf_storage, pickle_storage)


def pickle_to_csv(table_name: str, pickle_folder_path: str, csv_folder_path: str) -> None:
    pickle_storage = PickleStorage(pickle_folder_path)
    csv_storage = CsvStorage(csv_folder_path)
    convert_table_storage(table_name, pickle_storage, csv_storage)


def pickle_to_csv_all(pickle_folder_path: str, csv_folder_path: str) -> None:
    pickle_storage = PickleStorage(pickle_folder_path)
    csv_storage = CsvStorage(csv_folder_path)
    convert_all_tables_storage(pickle_storage, csv_storage)


def pickle_to_sql(table_name: str, pickle_folder_path: str, engine: Engine) -> None:
    pickle_storage = PickleStorage(pickle_folder_path)
    sql_storage = SqlStorage(engine)
    convert_table_storage(table_name, pickle_storage, sql_storage)


def pickle_to_sql_all(pickle_folder_path: str, engine: Engine) -> None:
    pickle_storage = PickleStorage(pickle_folder_path)
    sql_storage = SqlStorage(engine)
    convert_all_tables_storage(pickle_storage, sql_storage)


def pickle_to_hdf(table_name: str, pickle_folder_path: str, hdf_file_path: str) -> None:
    pickle_storage = PickleStorage(pickle_folder_path)
    hdf_storage = HdfStorage(hdf_file_path)
    convert_table_storage(table_name, pickle_storage, hdf_storage)


def pickle_to_hdf_all(pickle_folder_path: str, hdf_file_path: str) -> None:
    pickle_storage = PickleStorage(pickle_folder_path)
    hdf_storage = HdfStorage(hdf_file_path)
    convert_all_tables_storage(pickle_storage, hdf_storage)


import abc
from typing import Optional

from pandas import DataFrame


class IStorage(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __setitem__(self, key: str, other: DataFrame) -> None:
        """Store DataFrame."""
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, key: str) -> DataFrame:
        """Retrieve stored DataFrame."""
        raise NotImplementedError

    @abc.abstractmethod
    def __delitem__(self, key) -> None:
        """Delete stored DataFrame."""
        raise NotImplementedError
    
    @abc.abstractmethod
    def store(self, df: DataFrame, table_name: str, schema: Optional[str] = None) -> None:
        """Store DataFrame."""
        raise NotImplementedError
    
    @abc.abstractmethod
    def load(self, table_name: str, schema: Optional[str] = None) -> DataFrame:
        """Retrieve stored DataFrame."""
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, table_name: str, schema: Optional[str] = None) -> None:
        """Delete stored DataFrame."""
        raise NotImplementedError

    @abc.abstractmethod
    def table_names(self, schema: Optional[str] = None) -> list[str]:
        """Get list of stored table names."""
        raise NotImplementedError


    #__setitem__

    #__getitem__

    #__delitem__    

    #store

    #load

    #delete

    #load_metadata -Optional

    #table_names

    #metadata_names -Optional





"""
Store Pandas DataFrames in database tables.
Also saves the DataFrame metadata in database tables.

When stored table is loaded back into a DataFrame,
the indexes and datatypes are converted back from saved metadata.

Example
-------
import pandas as pd
import sqlalchemy as sa
import trashpandas as tp

df = pd.DataFrame({'name': ['Joe', 'Bob', 'John'], 'age': [23, 34, 44]})

# Create SqlStorage object with sqlite database connection string.
storage = tp.SqlStorage('sqlite:///test.db')
# or create an engine with SQLAlchemy and pass it into SqlStorage
engine = sa.create_engine('sqlite:///test.db')
storage = tp.SqlStorage(engine)

# Store DataFrame in database as table named 'people'
# and store metadata as table named '_people_metadata'
storage.store(df, 'people') 
# or assign DataFrame to item
storage['people'] = df

# Retrieve DataFrame from SqlStorage object.
df = storage.load('people')
# or use table name key
df = storage['people']

# Delete stored sql table using SqlStorage delete method.
storage.delete('people')
# or use del on table name key
del storage['people']

# Or use functions instead of SqlStorage class
tp.store_df_sql(df, 'people', engine)

df = tp.load_df_sql('people', engine)

tp.delete_table_sql('people', engine)
"""

from typing import Union, Optional

from pandas import DataFrame, read_sql_table
from sqlalchemy import inspect, Table, MetaData, create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.schema import DropTable

from trashpandas.interfaces import IStorage
from trashpandas.utils import cast_type, convert_meta_to_dict, df_metadata, name_no_names, unname_no_names


class SqlStorage(IStorage):
    def __init__(self, engine: Union[Engine, str]) -> None:
        """Takes SQLAlchemy Engine or database connection string."""
        if isinstance(engine, str):
            engine = create_engine(engine)
        self.engine = engine

    def __repr__(self):
        return f"SqlStorage('{self.engine.url}')"

    def __setitem__(self, key: str, other: DataFrame) -> None:
        """Store DataFrame and metadata in database."""
        self.store(other, key)

    def __getitem__(self, key: str) -> DataFrame:
        """Retrieve DataFrame from database."""
        return self.load(key)

    def __delitem__(self, key) -> None:
        """Delete DataFrame and metadata from database."""
        self.delete(key)
        
    def store(self, df: DataFrame, table_name: str, schema: Optional[str] = None) -> None:
        """Store DataFrame and metadata in database."""
        store_df_sql(df, table_name, self.engine, schema=schema)
    
    def load(self, table_name: str, schema: Optional[str] = None) -> DataFrame:
        """Retrieve DataFrame from database."""
        return load_df_sql(table_name, self.engine, schema=schema)

    def delete(self, table_name: str, schema: Optional[str] = None) -> None:
        """Delete DataFrame and metadata from database."""
        delete_table_sql(table_name, self.engine, schema)

    def load_metadata(self, table_name: str, schema: Optional[str] = None) -> DataFrame:
        """Retrieve DataFrame metadata from database."""
        return load_metadata_sql(table_name, self.engine, schema=schema)

    def table_names(self, schema: Optional[str] = None) -> list[str]:
        """Query database for list of non-metadata table names."""
        return table_names_sql(self.engine, schema=schema)

    def metadata_names(self, schema: Optional[str] = None) -> list[str]:
        """Query database for list of metadata table names."""
        return metadata_names_sql(self.engine, schema=schema)


def store_df_sql(df: DataFrame, table_name: str, engine: Engine, schema=None) -> None:
    """Store DataFrame and metadata in database."""
    df = df.copy()
    name_no_names(df)
    metadata = df_metadata(df)
    df.to_sql(table_name, engine, if_exists='replace', schema=schema)
    metadata.to_sql(f'_{table_name}_metadata', engine, if_exists='replace', index=False, schema=schema)


def load_df_sql(table_name: str, engine: Engine, schema=None) -> DataFrame:
    """Retrieve DataFrame from database."""
    meta_name = f'_{table_name}_metadata'

    if meta_name not in metadata_names_sql(engine, schema=schema):
        return _first_load_df_sql(table_name, engine, schema=schema)

    metadata = _read_cast_metadata_sql(meta_name, engine, schema)
    types = convert_meta_to_dict(metadata)
    indexes = list(metadata['column'][metadata['index']==True])
    df = read_sql_table(table_name, engine, schema=schema).astype(types).set_index(indexes)
    unname_no_names(df)
    return df


def delete_table_sql(table_name: str, engine: Engine, schema=None) -> None:
    """Delete DataFrame and metadata from database."""
    table = _get_table(table_name, engine, schema)
    metadata = _get_table(f'_{table_name}_metadata', engine, schema)
    engine.execute(DropTable(table))
    engine.execute(DropTable(metadata))


def load_metadata_sql(table_name: str, engine: Engine, schema=None) -> DataFrame:
    """Retrieve DataFrame metadata from database."""
    meta_name = f'_{table_name}_metadata'
    return _read_cast_metadata_sql(meta_name, engine, schema=schema)


def table_names_sql(engine: Engine, schema=None) -> list[str]:
    """Query database for list of non-metadata table names."""
    table_names = inspect(engine).get_table_names(schema=schema)
    return [name for name in table_names if '_metadata' not in name and name[0]!='_']


def metadata_names_sql(engine: Engine, schema=None) -> list[str]:
    """Query database for list of metadata table names."""
    table_names = inspect(engine).get_table_names(schema=schema)
    return [name for name in table_names if '_metadata' in name and name[0]=='_']


def _read_cast_metadata_sql(table_name: str, engine: Engine, schema=None) -> DataFrame:
    """Load metadata table and cast column datatypes column."""
    meta = read_sql_table(table_name, engine, schema=schema)
    meta['datatype'] = cast_type(meta['datatype'])
    return meta


def _first_load_df_sql(table_name: str, engine: Engine, schema=None) -> DataFrame:
    """Load a sql table that has no metadata stored, create and store metadata"""
    df = read_sql_table(table_name, engine, schema=schema)
    store_df_sql(df, table_name, engine)
    return df


def _get_table(table_name: str, engine: Engine, schema=None) -> Table:
    """Get SQLAlchemy Table mapped to database table."""
    metadata = MetaData(bind=engine, schema=schema)
    return Table(table_name,
                 metadata,
                 autoload=True,
                 schema=schema)


"""
Store Pandas DataFrames as CSV files in a specified directory path.
Also saves the DataFrame metadata in CSV files.

When saved CSV is loaded back into a DataFrame,
the indexes and datatypes are converted back from saved metadata.

Example
-------
import pandas as pd
import trashpandas as tp

df = pd.DataFrame({'name': ['Joe', 'Bob', 'John'], 'age': [23, 34, 44]})

# Create CsvStorage object with current directory path.
storage = tp.CsvStorage('')

# Store DataFrame as csv named 'people.csv'
# and store metadata as csv named '_people_metadata.csv'
storage.store(df, 'people') 
# or assign DataFrame to item
storage['people'] = df

# Retrieve DataFrame from CsvStorage object.
df = storage.load('people')
# or use table name key
df = storage['people']

# Delete stored csv files using CsvStorage delete method.
storage.delete('people')
# or use del on table name key
del storage['people']

# Or use functions instead of CsvStorage class
tp.store_df_csv(df, 'people', '')

df = tp.load_df_csv('people', '')

tp.delete_table_csv('people', '')
"""

import os

from pandas import DataFrame, read_csv

from trashpandas.interfaces import IStorage
from trashpandas.utils import cast_type, convert_meta_to_dict, df_metadata, name_no_names, unname_no_names


class CsvStorage(IStorage):
    def __init__(self, folder_path: str) -> None:
        """Takes folder path where DataFrames and metadata are stored as csv files."""
        self.path = folder_path

    def __repr__(self) -> str:
        return f"CsvStorage(path='{self.path}')"

    def __setitem__(self, key: str, other: DataFrame) -> None:
        """Store DataFrame and metadata as csv files."""
        self.store(other, key)

    def __getitem__(self, key: str) -> DataFrame:
        """Retrieve DataFrame from csv file."""
        return self.load(key)

    def __delitem__(self, key: str) -> None:
        """Delete DataFrame and metadata csv files."""
        self.delete(key)

    def store(self, df: DataFrame, table_name: str) -> None:
        """Store DataFrame and metadata as csv files."""
        store_df_csv(df, table_name, self.path)
    
    def load(self, table_name: str) -> DataFrame:
        """Retrieve DataFrame from csv file."""
        return load_df_csv(table_name, self.path)

    def delete(self, table_name: str) -> None:
        """Delete DataFrame and metadata csv files."""
        delete_table_csv(table_name, self.path)

    def load_metadata(self, table_name: str) -> DataFrame:
        """Retrieve DataFrame metadata from csv file."""
        return load_metadata_csv(table_name, self.path)

    def table_names(self) -> list[str]:
        """Get list of stored non-metadata table names."""
        return table_names_csv(self.path)

    def metdata_names(self) -> list[str]:
        """Get list of stored metadata table names."""
        return metadata_names_csv(self.path)


def store_df_csv(df: DataFrame, table_name: str, path: str) -> None:
    """Store DataFrame and metadata as csv files."""
    csv_path = _get_csv_path(table_name, path)
    metadata_path = _get_metadata_csv_path(table_name, path)
    df = df.copy()
    name_no_names(df)
    metadata = df_metadata(df)
    df.to_csv(csv_path)
    metadata.to_csv(metadata_path, index=False)


def load_df_csv(table_name: str, path: str) -> DataFrame:
    """Retrieve DataFrame from csv file."""
    csv_path = _get_csv_path(table_name, path)
    metadata_path = _get_metadata_csv_path(table_name, path)

    if not os.path.exists(metadata_path):
        return _first_load_df_csv(table_name, path)

    metadata = _read_cast_metadata_csv(table_name, path)
    types = convert_meta_to_dict(metadata)
    indexes = list(metadata['column'][metadata['index']==True])
    df = read_csv(csv_path).astype(types).set_index(indexes)
    unname_no_names(df)
    return df


def delete_table_csv(table_name: str, path: str) -> None:
    """Delete DataFrame and metadata csv files."""
    csv_path = _get_csv_path(table_name, path)
    metadata_path = _get_metadata_csv_path(table_name, path)
    os.remove(csv_path)
    os.remove(metadata_path)


def load_metadata_csv(table_name: str, path: str) -> DataFrame:
    meta_name = f'_{table_name}_metadata'
    return _read_cast_metadata_csv(meta_name, path)


def table_names_csv(path: str) -> list[str]:
    """Get list of stored non-metadata table names."""
    filenames = os.listdir(path)
    return [filename.split('.csv')[0] for filename in filenames
                if filename.endswith('.csv') and '_metadata' not in filename]


def metadata_names_csv(path: str) -> list[str]:
    """Get list of stored metadata table names."""
    filenames = os.listdir(path)
    return [filename.split('.csv')[0] for filename in filenames
                if filename.endswith('.csv') and '_metadata' in filename]


def _get_csv_path(table_name: str, path: str) -> str:
    """Return joined folder path and csv file name for DataFrame."""
    return os.path.join(path, f'{table_name}.csv')


def _get_metadata_csv_path(table_name: str, path: str) -> str:
    """Return joined folder path and csv file name for DataFrame metadata."""
    return os.path.join(path, f'_{table_name}_metadata.csv')


def _read_cast_metadata_csv(table_name: str, path: str) -> DataFrame:
    """Load metadata csv and cast column datatypes column."""
    metadata_path = _get_metadata_csv_path(table_name, path)
    meta = read_csv(metadata_path)
    meta['datatype'] = cast_type(meta['datatype'])
    return meta


def _first_load_df_csv(table_name: str, path: str) -> DataFrame:
    """Load a csv that has no metadata stored, create and store metadata"""
    csv_path = _get_csv_path(table_name, path)
    df = read_csv(csv_path)
    store_df_csv(df, table_name, path)
    return df


__version__ = '0.0.7'

import importlib.util

from trashpandas.sql import SqlStorage
from trashpandas.csv import CsvStorage
from trashpandas.pickle import PickleStorage

from trashpandas.sql import store_df_sql, load_df_sql, delete_table_sql, table_names_sql
from trashpandas.csv import store_df_csv, load_df_csv, delete_table_csv, table_names_csv
from trashpandas.pickle import store_df_pickle, load_df_pickle, delete_table_pickle, table_names_pickle

if importlib.util.find_spec('h5py'):
    from trashpandas.hdf5 import HdfStorage
    from trashpandas.hdf5 import store_df_hdf5, load_df_hdf5, delete_table_hdf5, table_names_hdf5


"""
Store Pandas DataFrames as pickle files in a specified directory path.
Also retriave Pandas DataFrames from pickle files.

Example
-------
import pandas as pd
import trashpandas as tp

df = pd.DataFrame({'name': ['Joe', 'Bob', 'John'], 'age': [23, 34, 44]})

# Create PickleStorage object with current directory path.
storage = tp.PickleStorage('')

# Store DataFrame as pickle file named 'people.pickle'
# and store metadata as pickle file named '_people_metadata.pickle'
storage.store(df, 'people') 
# or assign DataFrame to item
storage['people'] = df

# Retrieve DataFrame using PickleStorage load method.
df = storage.load('people')
# or use table name key
df = storage['people']

# Delete stored pickle files using PickleStorage delete method.
storage.delete('people')
# or use del on table name key
del storage['people']

# Or use functions instead of PickleStorage class
tp.store_df_pickle(df, 'people', '')

df = tp.load_df_pickle('people', '')

tp.delete_table_pickle('people', '')
"""

import os

from pandas import DataFrame, read_pickle

from trashpandas.interfaces import IStorage


class PickleStorage(IStorage):
    def __init__(self, folder_path: str, file_extension: str = '.pickle') -> None:
        """Takes folder path where DataFrames are stored as pickle files."""
        self.path = folder_path
        self.file_extension = file_extension

    def __repr__(self) -> str:
        return f"PickleStorage(path='{self.path}')"

    def __setitem__(self, key: str, other: DataFrame) -> None:
        """Store DataFrame pickle file."""
        self.store(other, key)

    def __getitem__(self, key: str) -> DataFrame:
        """Retrieve DataFrame from pickle file."""
        return self.load(key)

    def __delitem__(self, key: str) -> None:
        """Delete DataFrame pickle file."""
        self.delete(key)  

    def store(self, df: DataFrame, table_name: str) -> None:
        """Store DataFrame pickle file."""
        store_df_pickle(df, table_name, self.path, self.file_extension)

    def load(self, table_name: str) -> DataFrame:
        """Retrieve DataFrame from pickle file."""
        return load_df_pickle(table_name, self.path, self.file_extension)

    def delete(self, table_name: str) -> None:
        """Delete DataFrame pickle file."""
        delete_table_pickle(table_name, self.path, self.file_extension)

    def table_names(self) -> list[str]:
        """Get list of stored table names."""
        return table_names_pickle(self.path)


def store_df_pickle(df: DataFrame, table_name: str, path: str, file_extension: str = '.pickle') -> None:
    """Store DataFrame as pickle file."""
    pickle_path = _get_pickle_path(table_name, path, file_extension)
    df.to_pickle(pickle_path)


def load_df_pickle(table_name: str, path: str, file_extension: str = '.pickle') -> DataFrame:
    """Retrieve DataFrame from pickle file."""
    pickle_path = _get_pickle_path(table_name, path)
    return read_pickle(pickle_path)


def delete_table_pickle(table_name: str, path: str, file_extension: str = '.pickle') -> None:
    """Delete DataFrame pickle file."""
    pickle_path = _get_pickle_path(table_name, path)
    os.remove(pickle_path)


def table_names_pickle(path: str, file_extension: str = '.pickle') -> list[str]:
    """Get list of stored table names."""
    filenames = os.listdir(path)
    return [filename.split(file_extension)[0] for filename in filenames
                if filename.endswith(file_extension) and '_metadata' not in filename]


def _get_pickle_path(table_name: str, path: str, file_extension: str = '.pickle') -> str:
    """Return joined folder path and pickle file name for DataFrame."""
    return os.path.join(path, f'{table_name}{file_extension}')